@@ -6,6 +6,7 @@ import numpy as np
 from .optimization import bottleneck
 from .process import omp_set_nthreads
 from .params import Config
+from collections.abc import Mapping
 import random
 import os.path
 import itertools
@@ -16,8 +17,8 @@ __all__ = [
     'to_iterable', 'to_iterable_of_len', 'to_array', 'anynan', 'argsort', 'lexsort',
     'inverse_permutation', 'list_to_dict_list', 'atleast_2d_pad', 'transpose_list_list',
     'zero_if_close', 'pad', 'any_nonzero', 'add_with_None_0', 'chi_list', 'group_by_degeneracy',
-    'get_close', 'find_subclass', 'get_recursive', 'set_recursive', 'update_recursive', 'flatten',
-    'setup_logging', 'build_initial_state', 'setup_executable'
+    'get_close', 'find_subclass', 'get_recursive', 'set_recursive', 'update_recursive',
+    'merge_recursive', 'flatten', 'setup_logging', 'build_initial_state', 'setup_executable'
 ]
 
 
@@ -50,7 +51,7 @@ def to_iterable_of_len(a, L):
     return a
 
 
-def to_array(a, shape=(None, ), dtype=None):
+def to_array(a, shape=(None, ), dtype=None, allow_incommensurate=False):
     """Convert `a` to an numpy array and tile to matching dimension/shape.
 
     This function provides similar functionality as numpys broadcast, but not quite the same:
@@ -68,6 +69,9 @@ def to_array(a, shape=(None, ), dtype=None):
         For int entries, tile the array periodically to fit the len.
     dtype :
         Optionally specifies the data type.
+    allow_incommensurate : bool
+        Whether to raise an Error (``False``) or still tile to the desired shape and just "crop"
+        in the end.
 
     Returns
     -------
@@ -81,14 +85,24 @@ def to_array(a, shape=(None, ), dtype=None):
         else:  # extending dimensions is ambiguous, so we better raise an Error.
             raise ValueError("don't know how to cast `a` to required dimensions.")
     reps = [1] * a.ndim
+    need_crop = False
+    crop = [slice(None, None)] * a.ndim
     for i in range(a.ndim):
         if shape[i] is None:
             continue
-        if shape[i] % a.shape[i] != 0:
-            raise ValueError("incomensurate len for tiling from {0:d} to {1:d}".format(
-                a.shape[i], shape[i]))
         reps[i] = shape[i] // a.shape[i]
-    return np.tile(a, reps)
+        if shape[i] % a.shape[i] != 0:
+            if allow_incommensurate:
+                reps[i] = reps[i] +  1
+                crop[i] = slice(None, shape[i])
+                need_crop = True
+            else:
+                raise ValueError("incomensurate len for tiling from {0:d} to {1:d}".format(
+                    a.shape[i], shape[i]))
+    a = np.tile(a, reps)
+    if need_crop:
+        a = a[tuple(crop)]
+    return a
 
 
 if bottleneck is not None:
@@ -297,7 +311,7 @@ def zero_if_close(a, tol=1.e-15):
     a : ndarray
         numpy array to be rounded
     tol : float
-        the threashold which values to consider as '0'.
+        the threshold which values to consider as '0'.
     """
     if a.dtype == np.complex128 or a.dtype == np.complex64:
         ar = np.choose(np.abs(a.real) < tol, [a.real, np.zeros(a.shape)])
@@ -464,7 +478,6 @@ def group_by_degeneracy(E, *args, subset=None, cutoff=1.e-12):
     [(0, 2), (1, 4), (3, 6), (5,)]
     >>> group_by_degeneracy(E, k, cutoff=0.001)  # k and E need to be close
     [(0,), (1, 4), (2,), (3,), (5,), (6,)]
-
     """
     assert cutoff >= 0.
     E = np.asarray(E)
@@ -517,29 +530,55 @@ def find_subclass(base_class, subclass_name):
     ----------
     base_class : class
         The base class of which `subclass_name` is supposed to be a subclass.
-    subclass_name : str
-        Name of the class to be found.
+    subclass_name : str | type
+        The name (str) of the class to be found.
+        Alternatively, if a type is given, it is directly returned. In that case, a warning is
+        raised if it is not a subclass of `base_class`.
 
     Returns
     -------
-    subclass : None | class
+    subclass : class
         Class with name `subclass_name` which is a subclass of the `base_class`.
         None, if no subclass of the given name is found.
+
+    Raises
+    ------
+    ValueError: When no or multiple subclasses of `base_class` exists with that `subclass_name`.
     """
-    if base_class.__name__ == subclass_name:
-        return base_class
-    subclasses = base_class.__subclasses__()
-    for subcls in subclasses:
-        if subcls.__name__ == subclass_name:
-            return subcls
-    for subcls in subclasses:
-        found = find_subclass(subcls, subclass_name)  # recursion
-        if found is not None:
-            return found
-    return None
+    if not isinstance(subclass_name, str):
+        subclass = subclass_name
+        if not isinstance(subclass, type):
+            raise TypeError("expect a str or class for `subclass_name`, got " + repr(subclass))
+        if not issubclass(subclass, base_class):
+            # still allow it: might intend duck-typing. However, a warning should be raised!
+            warnings.warn(f"find_subclass: {subclass!r} is not subclass of {base_class!r}")
+        return subclass
+    found = set()
+    _find_subclass_recursion(base_class, subclass_name, found, set())
+    if len(found) == 0:
+        raise ValueError(f"No subclass of {base_class.__name__} called {subclass_name!r} defined. "
+                         "Maybe missing an import of a file with a custom class definition?")
+    elif len(found) == 1:
+        return found.pop()
+    else:
+        msg = f"There exist multiple subclasses of {base_class!r} with name {subclass_name!r}:"
+        raise ValueError('\n'.join([msg] + [repr(c) for c in found]))
 
 
-def get_recursive(nested_data, recursive_key, separator="/"):
+def _find_subclass_recursion(base_class, name_to_find, found, checked):
+    if base_class.__name__ == name_to_find:
+        found.add(base_class)
+    for subcls in base_class.__subclasses__():
+        if subcls in checked:
+            continue
+        _find_subclass_recursion(subcls, name_to_find, found, checked)
+        checked.add(subcls)
+
+
+_UNSET = object()  # sentinel
+
+
+def get_recursive(nested_data, recursive_key, separator=".", default=_UNSET):
     """Extract specific value from a nested data structure.
 
     Parameters
@@ -551,11 +590,14 @@ def get_recursive(nested_data, recursive_key, separator="/"):
         A leading `separator` is ignored.
     separator : str
         Separator for splitting `recursive_key` into subkeys.
+    default :
+        If not specified, the function raises a `KeyError` if the recursive_key is invalid.
+        If given, return this value when any of the nested dicts does not contain the subkey.
 
     Returns
     -------
     entry :
-        For example, ``recursive_key="/some/sub/key"`` will result in extracing
+        For example, ``recursive_key="some.sub.key"`` will result in extracing
         ``nested_data["some"]["sub"]["key"]``.
 
     See also
@@ -568,11 +610,13 @@ def get_recursive(nested_data, recursive_key, separator="/"):
     if not recursive_key:
         return nested_data  # return the original data if recursive_key is just "/"
     for subkey in recursive_key.split(separator):
+        if default is not _UNSET and subkey not in nested_data:
+            return default
         nested_data = nested_data[subkey]
     return nested_data
 
 
-def set_recursive(nested_data, recursive_key, value, separator="/", insert_dicts=False):
+def set_recursive(nested_data, recursive_key, value, separator=".", insert_dicts=False):
     """Same as :func:`get_recursive`, but set the data entry to `value`."""
     if recursive_key.startswith(separator):
         recursive_key = recursive_key[len(separator):]
@@ -584,7 +628,7 @@ def set_recursive(nested_data, recursive_key, value, separator="/", insert_dicts
     nested_data[subkeys[-1]] = value
 
 
-def update_recursive(nested_data, update_data, separator="/", insert_dicts=True):
+def update_recursive(nested_data, update_data, separator=".", insert_dicts=True):
     """Wrapper around :func:`set_recursive` to allow updating multiple values at once.
 
     It simply calls :func:`set_recursive` for each ``recursive_key, value in update_data.items()``.
@@ -593,7 +637,67 @@ def update_recursive(nested_data, update_data, separator="/", insert_dicts=True)
         set_recursive(nested_data, k, v, separator, insert_dicts)
 
 
-def flatten(mapping, separator='/'):
+def merge_recursive(*nested_data, conflict='error', path=None):
+    """Merge nested dictionaries `nested1` and `nested2`.
+
+    Parameters
+    ----------
+    *nested_data: dict of dict
+        Nested dictionaries that should be merged.
+    path: list of str
+        Path inside the nesting for useful error message.
+    conflict: "error" | "first" | "last"
+        How to handle conflicts: raise an error (if the values are different),
+        or just give priority to the first or last `nested_data` that still has a value,
+        even if they are different.
+
+    Returns
+    -------
+    merged: dict of dict
+        A single nested dictionary with the keys/values of the `nested_data` merged.
+        Dictionary values appearing in multiple of the `nested_data` get merged recursively.
+    """
+    if len(nested_data) == 0:
+        raise ValueError("need at least one nested_data")
+    elif len(nested_data) == 1:
+        return nested_data[0]
+    elif len(nested_data) > 2:
+        merged = nested_data[0]
+        for to_merge in nested_data[1:]:
+            merged = merge_recursive(merged, to_merge, conflict=conflict, path=path)
+        return merged
+    nested1, nested2 = nested_data
+    if path is None:
+        path = []
+    merged = nested1.copy()
+    for key, val2 in nested2.items():
+        if key in merged:
+            val1 = merged[key]
+            if isinstance(val1, Mapping) and isinstance(val2, Mapping):
+                merged[key] = merge_recursive(val1,
+                                              val2,
+                                              conflict=conflict,
+                                              path=path + [repr(key)])
+            else:
+                if conflict == 'error':
+                    if val1 != val2:
+                        path = ':'.join(path + [repr(key)])
+                        msg = '\n'.join([
+                            f"Conflict with different values at {path}; we got:",
+                            repr(val1),
+                            repr(val2)
+                        ])
+                        raise ValueError(msg)
+                elif conflict == 'first':
+                    pass
+                elif conflict == 'last':
+                    merged[key] = val2
+        else:
+            merged[key] = val2
+    return merged
+
+
+def flatten(mapping, separator='.'):
     """Obtain a flat dictionary with all key/value pairs of a nested data structure.
 
     Parameters
@@ -618,9 +722,9 @@ def flatten(mapping, separator='/'):
     >>> flat = flatten(sample_data)
     >>> for k in sorted(flat):
     ...     print(repr(k), ':', flat[k])
-    'some/nested/entry' : 100
-    'some/nested/structure' : 200
-    'some/subkey' : 10
+    'some.nested.entry' : 100
+    'some.nested.structure' : 200
+    'some.subkey' : 10
     'topentry' : 1
 
 
@@ -642,7 +746,22 @@ def flatten(mapping, separator='/'):
     return result
 
 
-def setup_logging(options={}, output_filename=None):
+#: default value for :cfg:option:`log.skip_setup`
+skip_logging_setup = False
+
+
+def setup_logging(options=None,
+                  output_filename=None,
+                  *,
+                  filename=None,
+                  to_stdout="INFO",
+                  to_file="INFO",
+                  format="%(levelname)-8s: %(message)s",
+                  datefmt=None,
+                  logger_levels={},
+                  dict_config=None,
+                  capture_warnings=None,
+                  skip_setup=None):
     """Configure the :mod:`logging` module.
 
     The default logging setup is given by the following equivalent `dict_config`
@@ -679,19 +798,23 @@ def setup_logging(options={}, output_filename=None):
         We **remove** any previously configured logging handlers.
         This is to handle the case when this function is called multiple times,
         e.g., because you run multiple :class:`~tenpy.simulations.simulation.Simulation`
-        classes sequentially.
+        classes sequentially (e.g., :func:`~tenpy.simulations.simulation.run_seq_simulations`).
+
+    .. deprecated :: 0.9.0
+        The arguments were previously collected in a dicitonary `options`.
+        Now they should be given directly as keyword arguments.
 
     Parameters
     ----------
-    options : dict
-        Parameters as described below.
+    **kwargs :
+        Keyword arguments as described in the options below.
     output_filename : None | str
-        The filename where results are saved. The `filename` for the log-file defaults to
-        this, but replecing the extension with ``.log``.
+        The filename for where results are saved. The :cfg:option:`log.filename` for the
+        log-file defaults to this, but replacing the extension with ``.log``.
 
     Options
     -------
-    .. cfg:config :: logging
+    .. cfg:config :: log
 
         skip_setup: bool
             If True, don't change anything in the logging setup; just return.
@@ -716,31 +839,38 @@ def setup_logging(options={}, output_filename=None):
             ``Simluation.logger`` class attribute. Hence, all messages from Simulation class
             methods calling ``self.logger.info(...)`` will be affected by that.
         format : str
-            Formatting string, `fmt` argument of :class:`logging.config.Formatter`.
+            Formatting string, `fmt` argument of :class:`logging.Formatter`.
+            You can for example use ``"{loglevel:.4s} {asctime} {message}"`` to include the time
+            stamp of each message into the log - this is usefull to get an idea where code hangs.
+            Find
+            `allowed keys <https://docs.python.org/3/library/logging.html#logrecord-attributes>`_
+            here. The style of the formatter is chosen depending on whether the format string
+            contains ``'%' '{' '$'``, respectively.
+        datefmt : str
+            Formatting string for the `asctime` key in the `format`, e.g. ``"%Y-%m-%d %H:%M:%S"``,
+            see :meth:`logging.Formatter.formatTime`.
         dict_config : dict
-            Alternatively, a full configuration dictionary for :mod:`logging.config.dictConfig`.
+            Alternatively, a full configuration dictionary for :func:`logging.config.dictConfig`.
             If used, all other options except `skip_setup` and `capture_warnings` are ignored.
         capture_warnings : bool
             Whether to call :func:`logging.captureWarnings` to include the warnings into the log.
     """
     import logging.config
-    if output_filename is None:
-        default_log_fn = None
-    else:
-        root, ext = os.path.splitext(output_filename)
-        assert ext != 'log'
-        default_log_fn = root + '.log'
-    log_fn = options.get('filename', default_log_fn)
-    to_stdout = options.get('to_stdout', "INFO")
-    to_file = options.get('to_file', "INFO")
-    log_format = options.get('format', "%(levelname)-8s: %(message)s")
-    logger_levels = options.get('logger_levels', {})
-    conf = options.get('dict_config', None)
-    capture_warnings = options.get('capture_warnings', conf is not None
-                                   or bool(to_stdout or to_file))
-    if options.get('skip_setup', False):
+    if options is not None:
+        warnings.warn("Give logging parameters directly as keyword arguments!", FutureWarning, 2)
+        locals().update(**options)
+    if filename is None:
+        if output_filename is not None:
+            root, ext = os.path.splitext(output_filename)
+            assert ext != '.log'
+            filename = root + '.log'
+    if capture_warnings is None:
+        capture_warnings = dict_config is not None or to_stdout or to_file
+    if skip_setup is None:
+        skip_setup = skip_logging_setup
+    if skip_setup:
         return
-    if conf is None:
+    if dict_config is None:
         handlers = {}
         if to_stdout:
             handlers['to_stdout'] = {
@@ -749,23 +879,24 @@ def setup_logging(options={}, output_filename=None):
                 'formatter': 'custom',
                 'stream': 'ext://sys.stdout',
             }
-        if to_file and log_fn is not None:
+        if to_file and filename is not None:
             handlers['to_file'] = {
                 'class': 'logging.FileHandler',
                 'level': to_file,
                 'formatter': 'custom',
-                'filename': log_fn,
+                'filename': filename,
                 'mode': 'a',
             }
             if not to_stdout:
                 cwd = os.getcwd()
-                print(f"now logging to {cwd!r}/{log_fn!r}")
-        conf = {
+                print(f"now logging to {cwd!s}/{filename!s}")
+        dict_config = {
             'version': 1,  # mandatory
             'disable_existing_loggers': False,
             'formatters': {
                 'custom': {
-                    'format': log_format
+                    'format': format,
+                    'datefmt': datefmt
                 }
             },
             'handlers': handlers,
@@ -775,15 +906,23 @@ def setup_logging(options={}, output_filename=None):
             },
             'loggers': {},
         }
+        if '%' not in format:
+            if '{' in format:
+                assert '$' not in format
+                style = '{'
+            else:
+                assert '$' in format
+                style = '$'
+            dict_config['formatters']['custom']['style'] = style
         for name, level in logger_levels.items():
             if name == 'root':
-                conf['root']['level'] = level
+                dict_config['root']['level'] = level
             else:
-                conf['loggers'].setdefault(name, {})['level'] = level
+                dict_config['loggers'].setdefault(name, {})['level'] = level
     else:
-        conf.setdefault('disable_existing_loggers', False)
+        dict_config.setdefault('disable_existing_loggers', False)
     # note: dictConfig cleans up previously existing handlers etc
-    logging.config.dictConfig(conf)
+    logging.config.dictConfig(dict_config)
     if capture_warnings:
         logging.captureWarnings(True)
 
